@@ -5,36 +5,38 @@ things in hidden_things right now:
     - ModelStructure
 '''
 
+import z3
 import sys
-
 from contextlib import redirect_stdout
-
 import time
-
 from functools import reduce
-
 from string import Template
-
 from z3 import (
     And,
     ArrayRef,
     BitVecSort,
+    BitVecVal,
     EmptySet,
+    ExprRef,
+    IntVal,
     IsMember,
     Not,
-    BitVecVal,
     SetAdd,
     Solver,
     sat,
     simplify,
 )
 
-from .utils import (
-    bitvec_to_substates,
-    int_to_binary,
-    not_implemented_string,
-    pretty_set_print,
-)
+# Try local imports first (for development)
+try:
+    from src.model_checker.utils import (
+        not_implemented_string,
+    )
+except ImportError:
+    # Fall back to installed package imports
+    from model_checker.utils import (
+        not_implemented_string,
+    )
 
 inputs_template = Template(
 '''Z3 run time: ${z3_model_runtime} seconds
@@ -78,20 +80,27 @@ class SemanticDefaults:
     """Includes default attributes and methods to be inherited by a semantics
     including frame constraints, truth and falsity, and logical consequence."""
 
-    def __init__(self, N):
+    def __init__(self, combined_settings):
 
         # Store the name
         self.name = self.__class__.__name__
 
-        # Store the number of states
-        self.N = N
+        # Define all states and top and bottom if N is specified
+        if 'N' in combined_settings.keys():
+            self.N = combined_settings['N']
+            max_value = (1 << self.N) - 1 # NOTE: faster than 2**self.N - 1
+            self.full_bit = BitVecVal(max_value, self.N)
+            self.null_bit = BitVecVal(0, self.N)
+            self.all_bits = [BitVecVal(i, self.N) for i in range(1 << self.N)]
+
+        # Define all times between 0 and M inclusive
+        if 'M' in combined_settings.keys():
+            self.M = combined_settings['M']
+            self.all_times = [IntVal(i) for i in range(self.M)]
+
+        # Define main_point
+        self.main_point = None
 
-        # Define top and bottom states
-        max_value = (1 << self.N) - 1 # NOTE: faster than 2**self.N - 1
-        self.full_bit = BitVecVal(max_value, self.N)
-        self.null_bit = BitVecVal(0, self.N)
-        self.all_bits = [BitVecVal(i, self.N) for i in range(1 << self.N)]
-        
         # Define the frame constraints
         self.frame_constraints = None
 
@@ -163,49 +172,33 @@ class PropositionDefaults:
         if self.__class__ == PropositionDefaults:
             raise NotImplementedError(not_implemented_string(self.__class__.__name__))
 
-        # Store values from sentence
-        self.name = sentence.name
-        self.operator = sentence.operator
-        self.arguments = sentence.arguments
-        self.sentence_letter = sentence.sentence_letter
+        # Validate model_structure
+        if not hasattr(model_structure, 'semantics'):
+            raise TypeError(
+                f"Expected model_structure to be a ModelStructure object, got {type(model_structure)}. "
+                "Make sure you're passing the correct model_structure object when creating propositions."
+            )
 
-        # Store values from model_structure argument
+        # Store inputs
+        self.sentence = sentence
         self.model_structure = model_structure
-        self.N = self.model_structure.N
-        self.model_constraints = self.model_structure.model_constraints
+
+        # Store values from model_structure
+        self.N = self.model_structure.semantics.N
+        self.main_point = self.model_structure.main_point
+
+        # Store values from sentence
+        self.name = self.sentence.name
+        self.operator = self.sentence.operator
+        self.arguments = self.sentence.arguments
+        self.sentence_letter = self.sentence.sentence_letter
 
         # Store values from model_constraints
+        self.model_constraints = self.model_structure.model_constraints
         self.semantics = self.model_constraints.semantics
         self.sentence_letters = self.model_constraints.sentence_letters
         self.settings = self.model_constraints.settings
 
-        # Set defaults for verifiers and falsifiers (important they are lists)
-        self.verifiers, self.falsifiers = [], []
-
-    # NOTE: is responsive to unilateral/bilateral props, so long as
-    # falsifiers, if there are any, are _sets_; the default is a list,
-    # so if it is a list, it is because the defaults are unchanged, meaning
-    # the proposition is unilateral (a prop with no falsifiers but within
-    # a bilateral system would have an empty set as falsifiers, not the
-    # default empty list)
-    def __repr__(self):
-        N = self.model_structure.model_constraints.semantics.N
-        possible = self.model_structure.model_constraints.semantics.possible
-        z3_model = self.model_structure.z3_model
-        ver_states = {
-            bitvec_to_substates(bit, N)
-            for bit in self.verifiers
-            if z3_model.evaluate(possible(bit)) or self.settings['print_impossible']
-        }
-        if isinstance(self.falsifiers, set): # because default is an empty list
-            fal_states = {
-                bitvec_to_substates(bit, N)
-                for bit in self.falsifiers
-                if z3_model.evaluate(possible(bit)) or self.settings['print_impossible']
-            }
-            return f"< {pretty_set_print(ver_states)}, {pretty_set_print(fal_states)} >"
-        return pretty_set_print(ver_states)
-
     def __hash__(self):
         return hash(self.name)
 
@@ -223,13 +216,13 @@ class PropositionDefaults:
         if indent_num == 1:
             FULL, PART = (GREEN, GREEN) if truth_value else (RED, RED)
             if truth_value is None:
-                # world_state = bitvec_to_substates(eval_world, N)
                 print(
                     f"\n{RED}WARNING:{RESET}"
                     f"{name} is neither true nor false at {world_state}.\n"
                 )
         return RESET, FULL, PART
 
+
 class ModelConstraints:
     """Takes semantics and proposition_class as arguments to build generate
     and storing all Z3 constraints. This class is passed to ModelStructure."""
@@ -247,11 +240,13 @@ class ModelConstraints:
         self.semantics = semantics
         self.proposition_class = proposition_class
         self.settings = settings
+        # self.old_z3_models = None
 
         # Store syntax values
         self.premises = self.syntax.premises
         self.conclusions = self.syntax.conclusions
-        self.sentence_letters = self.syntax.sentence_letters
+
+        self.sentence_letters = self._load_sentence_letters()
 
         # Store operator dictionary
         self.operators = self.copy_dictionary(self.syntax.operator_collection)
@@ -266,21 +261,15 @@ class ModelConstraints:
             for sentence_letter in self.sentence_letters
             for constraint in self.proposition_class.proposition_constraints(
                 self,
-                sentence_letter.sentence_letter,
+                sentence_letter,
             )
         ]
         self.premise_constraints = [
-            self.semantics.premise_behavior(
-                premise,
-                self.semantics.main_world,
-            )
+            self.semantics.premise_behavior(premise)
             for premise in self.premises
         ]
         self.conclusion_constraints = [
-            self.semantics.conclusion_behavior(
-                conclusion,
-                self.semantics.main_world,
-            )
+            self.semantics.conclusion_behavior(conclusion)
             for conclusion in self.conclusions
         ]
         self.all_constraints = (
@@ -296,6 +285,21 @@ class ModelConstraints:
         conclusions = list(self.syntax.conclusions)
         return f"ModelConstraints for premises {premises} and conclusions {conclusions}"
 
+    # def _store_z3_model(self, old_z3_model):
+    #     if self.old_z3_models is None:
+    #         self.old_z3_models = [old_z3_model]
+    #     else:
+    #         self.old_z3_models.append(old_z3_model)
+
+    def _load_sentence_letters(self):
+        sentence_letters = []
+        for packed_letter in self.syntax.sentence_letters:
+            unpacked_letter = packed_letter.sentence_letter
+            if not isinstance(unpacked_letter, ExprRef):
+                raise TypeError("The sentence letter {letter} is not of type z3.ExprRef")
+            sentence_letters.append(unpacked_letter) 
+        return sentence_letters
+
     def copy_dictionary(self, operator_collection):
         """Copies the operator_dictionary from operator_collection."""
         return {
@@ -337,63 +341,12 @@ class ModelConstraints:
             for index, sent in enumerate(conclusions, start=start_con_num):
                 print(f"{index}. {sent}", file=output)
 
-class ModelStructure:
+
+# TODO: expose elements that need to change to accommodate bimodal logic
+class ModelDefaults:
     """Solves and stores the Z3 model for an instance of ModelSetup."""
 
-    def __init__(self, model_constraints, max_time=1):
-        self.constraint_dict = {} # hopefully temporary, for unsat_core
-
-        # Store arguments
-        self.model_constraints = model_constraints
-        self.max_time = max_time
-
-        # Store from model_constraint.syntax
-        self.syntax = self.model_constraints.syntax
-        self.start_time = self.syntax.start_time
-        self.premises = self.syntax.premises
-        self.conclusions = self.syntax.conclusions
-        self.sentence_letters = self.syntax.sentence_letters
-
-        # Store from model_constraint.semantics
-        self.semantics = self.model_constraints.semantics
-        self.main_world = self.semantics.main_world
-        self.all_bits = self.semantics.all_bits
-        self.N = self.semantics.N
-
-        # Store from model_constraint
-        self.proposition_class = self.model_constraints.proposition_class
-        self.settings = self.model_constraints.settings
-
-        # Solve Z3 constraints and store values
-        timeout, z3_model, z3_model_status, z3_model_runtime = self.solve(
-            self.model_constraints,
-            self.max_time,
-        )
-        self.timeout = timeout
-        self.z3_model = z3_model if z3_model_status else None
-        self.unsat_core = z3_model if not z3_model_status else None
-        self.z3_model_status = z3_model_status
-        self.z3_model_runtime = z3_model_runtime
-
-        # Store possible_bits, world_bits, and main_world from the Z3 model
-        if not self.z3_model_status:
-            self.poss_bits, self.world_bits, self.main_world = None, None, None
-            return
-        self.poss_bits = [
-            bit
-            for bit in self.all_bits
-            if bool(self.z3_model.evaluate(self.semantics.possible(bit)))
-            # LINTER: cannot access attribute "evaluate" for class "AstVector"
-        ]
-        self.world_bits = [
-            bit
-            for bit in self.poss_bits
-            if bool(self.z3_model.evaluate(self.semantics.is_world(bit)))
-            # LINTER: cannot access attribute "evaluate" for class "AstVector"
-        ]
-        if not self.z3_model is None:
-            self.main_world = self.z3_model[self.main_world]
-
+    def __init__(self, model_constraints, settings):
         # Define ANSI color codes
         self.COLORS = {
             "default": "\033[37m",  # WHITE
@@ -405,43 +358,140 @@ class ModelStructure:
         self.RESET = "\033[0m"
         self.WHITE = self.COLORS["default"]
 
+        # TODO: is this still needed?
+        self.constraint_dict = {} # hopefully temporary, for unsat_core
+
+        # Store arguments
+        self.model_constraints = model_constraints
+        self.settings = settings
+        self.max_time = self.settings["max_time"]
+        self.expectation = self.settings["expectation"]
+
+        # Store from model_constraints.semantics
+        self.semantics = self.model_constraints.semantics
+        self.main_point = self.semantics.main_point
+        self.all_bits = self.semantics.all_bits # TODO: move?
+        self.N = self.semantics.N
+
+        # Store from model_constraint.syntax
+        self.syntax = self.model_constraints.syntax
+        self.start_time = self.syntax.start_time
+        self.premises = self.syntax.premises
+        self.conclusions = self.syntax.conclusions
+        self.sentence_letters = self.syntax.sentence_letters
+
+        # Store from model_constraint
+        self.proposition_class = self.model_constraints.proposition_class
+        self.settings = self.model_constraints.settings
+
+        # Initialize Z3 attributes
+        self.solver = None # TODO: still needed?
+        self.timeout = None
+        self.z3_model = None
+        self.unsat_core = None
+        self.z3_model_status = None
+        self.z3_model_runtime = None
+
+        # Solve Z3 constraints and store results
+        solver_results = self.solve(self.model_constraints, self.max_time)
+        self._process_solver_results(solver_results)
+
+        # Exit early if no valid model was found
+        if self.timeout or self.z3_model is None:
+            return
+
         # Recursively update propositions
         self.interpret(self.premises + self.conclusions)
 
-    def solve(self, model_constraints, max_time):
-        solver = Solver()
-        # Track each constraint with a unique name
-        fc, mc, pc, cc = (
-            model_constraints.frame_constraints,
-            model_constraints.model_constraints,
-            model_constraints.premise_constraints,
-            model_constraints.conclusion_constraints
-        )
-        for c_group, c_group_name in [
-            (fc, "frame"),
-            (mc, "model"),
-            (pc, "premises"),
-            (cc, "conclusions")
-        ]:
-            for ix, c in enumerate(c_group):
-                c_id = f"{c_group_name}{ix+1}"
-                solver.assert_and_track(c, c_id)
-                self.constraint_dict[c_id] = c
+    def _process_solver_results(self, solver_results):
+        """Process and store the results from the Z3 solver.
+        
+        Args:
+            solver_results: Tuple containing (timeout, z3_model, z3_model_status, z3_model_runtime)
+        """
+        timeout, z3_model, z3_model_status, z3_model_runtime = solver_results
+        
+        self.timeout = timeout
+        self.z3_model_status = z3_model_status
+        self.z3_model_runtime = z3_model_runtime
+        
+        # Store either the model or unsat core based on status
+        if z3_model_status:
+            self.z3_model = z3_model
+        else:
+            self.unsat_core = z3_model
 
-        solver.set("timeout", int(max_time * 1000))  # time in seconds
+    def _setup_solver(self, model_constraints):
+        """Initialize Z3 solver and add base constraints."""
+        solver = Solver()
+        constraint_groups = [
+            (model_constraints.frame_constraints, "frame"),
+            (model_constraints.model_constraints, "model"), 
+            (model_constraints.premise_constraints, "premises"),
+            (model_constraints.conclusion_constraints, "conclusions")
+        ]
+        
+        for constraints, group_name in constraint_groups:
+            for ix, constraint in enumerate(constraints):
+                c_id = f"{group_name}{ix+1}"
+                solver.assert_and_track(constraint, c_id)
+                self.constraint_dict[c_id] = constraint
+                
+        return solver
+
+    def _create_result(self, is_timeout, model_or_core, is_satisfiable, start_time):
+        """Create a standardized result tuple."""
+        runtime = round(time.time() - start_time, 4)
+        return is_timeout, model_or_core, is_satisfiable, runtime
+
+    def solve(self, model_constraints, max_time):
+        """Solve the model constraints and return the results."""
         try:
-            model_start = time.time()  # start benchmark timer
-            result = solver.check()
-            model_end = time.time()  # end benchmark timer
-            model_runtime = round(model_end - model_start, 4)
+            self.solver = self._setup_solver(model_constraints)
+
+            # Set timeout and solve
+            self.solver.set("timeout", int(max_time * 1000))
+            start_time = time.time()
+            result = self.solver.check()
+            
+            # Handle different solver outcomes
             if result == sat:
-                return False, solver.model(), True, model_runtime
-            if solver.reason_unknown() == "timeout":
-                return True, None, False, model_runtime
-            return False, solver.unsat_core(), False, model_runtime
-        except RuntimeError as e:  # Handle unexpected exceptions
+                return self._create_result(False, self.solver.model(), True, start_time)
+            
+            if self.solver.reason_unknown() == "timeout":
+                return self._create_result(True, None, False, start_time)
+            
+            return self._create_result(False, self.solver.unsat_core(), False, start_time)
+            
+        except RuntimeError as e:
             print(f"An error occurred while running `solve_constraints()`: {e}")
             return True, None, False, None
+    
+    def re_solve(self):
+        """Re-solve the existing model constraints and return the results."""
+        try:
+            assert self.solver is not None
+            # Re-apply timeout setting
+            self.solver.set("timeout", int(self.max_time * 1000))
+            
+            start_time = time.time()
+            result = self.solver.check()
+            
+            # Handle different solver outcomes
+            if result == sat:
+                return self._create_result(False, self.solver.model(), True, start_time)
+            
+            if self.solver.reason_unknown() == "timeout":
+                return self._create_result(True, None, False, start_time)
+            
+            return self._create_result(False, self.solver.unsat_core(), False, start_time)
+            
+        except RuntimeError as e:
+            print(f"An error occurred while running `re_solve()`: {e}")
+            return True, None, False, None
+
+    def check_result(self):
+        return self.z3_model_status == self.settings["expectation"]
 
     def interpret(self, sentences):
         """Updates each instance of Sentence in sentences by adding the
@@ -454,20 +504,6 @@ class ModelStructure:
                 self.interpret(sent_obj.arguments)
             sent_obj.update_proposition(self)
 
-    def print_evaluation(self, output=sys.__stdout__):
-        """print the evaluation world and all sentences letters that true/false
-        in that world"""
-        BLUE = ""
-        RESET = ""
-        main_world = self.main_world
-        if output is sys.__stdout__:
-            BLUE = "\033[34m"
-            RESET = "\033[0m"
-        print(
-            f"\nThe evaluation world is: {BLUE}{bitvec_to_substates(main_world, self.N)}{RESET}\n",
-            file=output,
-        )
-
     def print_grouped_constraints(self, output=sys.__stdout__):
         """Prints constraints organized by their groups (frame, model, premises, conclusions)"""
         groups = {
@@ -479,13 +515,13 @@ class ModelStructure:
         
         # Get the relevant constraints based on model status
         if self.z3_model:
-            print("\nSATISFIABLE CONSTRAINTS:", file=output)
+            print("SATISFIABLE CONSTRAINTS:", file=output)
             constraints = self.model_constraints.all_constraints
         elif self.unsat_core is not None:
-            print("\nUNSATISFIABLE CORE CONSTRAINTS:", file=output)
+            print("UNSATISFIABLE CORE CONSTRAINTS:", file=output)
             constraints = [self.constraint_dict[str(c)] for c in self.unsat_core]
         else:
-            print("\nNO CONSTRAINTS AVAILABLE", file=output)
+            print("NO CONSTRAINTS AVAILABLE", file=output)
             constraints = []
             
         # Print summary of constraint groups
@@ -501,7 +537,6 @@ class ModelStructure:
         
         # Organize constraints into groups
         for constraint in constraints:
-            constraint_str = str(constraint)
             if constraint in self.model_constraints.frame_constraints:
                 groups["FRAME"].append(constraint)
             elif constraint in self.model_constraints.model_constraints:
@@ -527,24 +562,6 @@ class ModelStructure:
         for index, con in enumerate(constraints, start=1):
             print(f"{index}. {con}\n", file=output)
 
-    def print_to(self, default_settings, example_name, theory_name, print_constraints=None, output=sys.__stdout__):
-        """append all elements of the model to the file provided
-        
-        Args:
-            print_constraints: Whether to print constraints. Defaults to value in settings.
-            output: Output stream to print to. Defaults to sys.stdout.
-        """
-        if print_constraints is None:
-            print_constraints = self.settings["print_constraints"]
-        if self.timeout:
-            print(f"TIMEOUT: {self.timeout}")
-            print(f"No model for example {example_name} found before timeout.", file=output)
-            print(f"Try increasing max_time > {self.max_time}.\n", file=output)
-            return
-        self.print_all(default_settings, example_name, theory_name, output)
-        if print_constraints and self.unsat_core is not None:
-            self.print_grouped_constraints(output)
-
     def build_test_file(self, output):
         """generates a test file from input to be saved"""
 
@@ -557,58 +574,14 @@ class ModelStructure:
         inputs_content = inputs_template.substitute(inputs_data)
         print(inputs_content, file=output)
 
-
-    def save_to(self, example_name, theory_name, include_constraints, output):
-        """append all elements of the model to the file provided"""
-        constraints = self.model_constraints.all_constraints
-        self.print_all(example_name, theory_name, output)
-        self.build_test_file(output)
-        if include_constraints:
-            print("# Satisfiable constraints", file=output)
-            print(f"all_constraints = {constraints}", file=output)
-
-    def print_states(self, output=sys.__stdout__):
-        """Print all fusions of atomic states in the model."""
-
-        def binary_bitvector(bit):
-            return (
-                bit.sexpr()
-                if self.N % 4 != 0
-                else int_to_binary(int(bit.sexpr()[2:], 16), self.N)
-            )
-        
-        def format_state(bin_rep, state, color, label=""):
-            """Helper function to format and print a state."""
-            label_str = f" ({label})" if label else ""
-            use_colors = output is sys.__stdout__
-            if use_colors:
-                print(f"  {self.WHITE}{bin_rep} = {color}{state}{label_str}{self.RESET}", file=output)
-            else:
-                print(f"  {bin_rep} = {state}{label_str}", file=output)
-        
-        # Print formatted state space
-        print("State Space:", file=output)
-        for bit in self.all_bits:
-            state = bitvec_to_substates(bit, self.N)
-            bin_rep = binary_bitvector(bit)
-            if bit == 0:
-                format_state(bin_rep, state, self.COLORS["initial"])
-            elif bit in self.world_bits:
-                format_state(bin_rep, state, self.COLORS["world"], "world")
-            elif bit in self.poss_bits:
-                format_state(bin_rep, state, self.COLORS["possible"])
-            elif self.settings['print_impossible']:
-                format_state(bin_rep, state, self.COLORS["impossible"], "impossible")
-
-    def recursive_print(self, sentence, eval_world, indent_num, use_colors):
+    def recursive_print(self, sentence, eval_point, indent_num, use_colors):
         if indent_num == 2:  # NOTE: otherwise second lines don't indent
             indent_num += 1
         if sentence.sentence_letter is not None:  # Print sentence letter
-            # print(f"OUTPUT REC PRINT: {output is sys.__stdout__}")
-            sentence.proposition.print_proposition(eval_world, indent_num, use_colors)
+            sentence.proposition.print_proposition(eval_point, indent_num, use_colors)
             return
         operator = sentence.original_operator
-        operator.print_method(sentence, eval_world, indent_num, use_colors)  # Print complex sentence
+        operator.print_method(sentence, eval_point, indent_num, use_colors)  # Print complex sentence
 
     def print_input_sentences(self, output):
         """Prints the interpreted premises and conclusions, leveraging recursive_print."""
@@ -621,9 +594,8 @@ class ModelStructure:
                 for index, sentence in enumerate(sentences, start=start_index):
                     print(f"{index}.", end="", file=output)
                     with redirect_stdout(destination):
-                        # print(f"OUTPUT PRINT_SENT {destination} is STD: {output is sys.__stdout__}")
                         use_colors = output is sys.__stdout__
-                        self.recursive_print(sentence, self.main_world, 1, use_colors)
+                        self.recursive_print(sentence, self.main_point, 1, use_colors)
                         print(file=output)
         
         start_index = 1
@@ -702,18 +674,3 @@ class ModelStructure:
         print(f"\nZ3 Run Time: {self.z3_model_runtime} seconds", file=output)
         print(f"\n{'='*40}\n", file=output)
 
-    def print_all(self, default_settings, example_name, theory_name, output=sys.__stdout__):
-        """prints states, sentence letters evaluated at the designated world and
-        recursively prints each sentence and its parts"""
-        model_status = self.z3_model_status
-        self.print_info(model_status, default_settings, example_name, theory_name, output)
-        if model_status:
-            self.print_states(output)
-            self.print_evaluation(output)
-            self.print_input_sentences(output)
-            self.print_model(output)
-            if output is sys.__stdout__:
-                total_time = round(time.time() - self.start_time, 4) 
-                print(f"Total Run Time: {total_time} seconds\n", file=output)
-                print(f"{'='*40}", file=output)
-            return
